# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    calc_budget_and_residual.py                        :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/01/31 20:15:59 by daniloceano       #+#    #+#              #
#    Updated: 2024/01/03 22:33:44 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
Budget Difference and Residual Calculations for the Lorenz Energy Cycle.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import logging

import numpy as np
import pandas as pd


def calc_budget_diff(df: pd.DataFrame, dates: np.ndarray, app_logger: logging.Logger):
    """
    Estimate budget values for energy terms using finite differences method.

    Args:
        df (DataFrame): DataFrame containing energy terms.
        dates (Array-like): Array of datetime objects representing time points.

    Returns:
        DataFrame: Updated DataFrame with budget values.
    """
    app_logger.debug("Estimating budget values using finite differences...")

    dt = float((dates[1] - dates[0]) / np.timedelta64(1, "s"))
    energy_terms = ["Az", "Ae", "Kz", "Ke"]

    try:
        for term in energy_terms:
            df[f"∂{term}/∂t (finite diff.)"] = np.gradient(df[term], dt)
    except Exception as e:
        app_logger.error(f"Error in calc_budget_diff: {e}")
        raise

    app_logger.debug("Done.")
    return df


def calc_budget_diff_4th(
    df: pd.DataFrame, time: np.ndarray, app_logger: logging.Logger
):
    """
    Estimate budget values for energy terms using 4th order finite differences.

    Args:
        df (DataFrame): DataFrame containing energy terms.
        time (Array-like): Array of datetime objects representing time points.

    Returns:
        DataFrame: Updated DataFrame with budget values.
    """
    app_logger.debug("Estimating budget values using 4th order finite differences...")
    dt = float((time[1] - time[0]) / np.timedelta64(1, "s"))
    energy_terms = ["Az", "Ae", "Kz", "Ke"]

    try:
        for term in energy_terms:
            df = _apply_4th_order_diff(df, term, dt)
    except Exception as e:
        app_logger.error(f"Error in calc_budget_diff_4th: {e}")
        raise

    app_logger.debug("Done.")
    return df


def _apply_4th_order_diff(df, term, dt):
    """
    Apply 4th order finite difference to a specific energy term.

    Args:
        df (DataFrame): DataFrame containing energy terms.
        term (str): The energy term to apply the finite difference.
        dt (float): Time difference in seconds.

    Returns:
        DataFrame: Updated DataFrame with the specific term calculated.
    """
    forward = (df[term].iloc[1] - df[term].iloc[0]) / dt
    central_second = (df[term].iloc[2] - df[term].iloc[0]) / (2 * dt)
    central_penultimate = (df[term].iloc[-1] - df[term].iloc[-3]) / (2 * dt)
    fourth_order = _compute_4th_order_terms(df[term], dt)
    backward = (df[term].iloc[-1] - df[term].iloc[-2]) / dt

    df[f"∂{term}/∂t (finite diff.)"] = (
        [forward, central_second] + fourth_order + [central_penultimate, backward]
    )
    return df


def _compute_4th_order_terms(series, dt):
    """
    Compute 4th order finite difference terms.

    Args:
        series (Series): Pandas Series containing the specific energy term.
        dt (float): Time difference in seconds.

    Returns:
        list: List of calculated 4th order terms.
    """
    fourth_order1 = (
        (4 / 3) * (series.iloc[3:-1].values - series.iloc[1:-3].values) / (2 * dt)
    )
    fourth_order2 = (
        (1 / 3) * (series.iloc[4:].values - series.iloc[:-4].values) / (4 * dt)
    )
    return list(fourth_order1 - fourth_order2)


def calc_residuals(df: pd.DataFrame, app_logger: logging.Logger):
    """
    Compute the residuals RGz, RKz, RGe, and RKe using estimated budget terms.

    Args:
        df (DataFrame): DataFrame containing budget and conversion terms.

    Returns:
        DataFrame: Updated DataFrame with residuals.
    """
    app_logger.debug("Estimating residuals...")

    try:
        df["RGz"] = df["∂Az/∂t (finite diff.)"] + df["Cz"] + df["Ca"] - df["BAz"]
        df["RKz"] = df["∂Kz/∂t (finite diff.)"] - df["Cz"] - df["Ck"] - df["BKz"]
        df["RGe"] = df["∂Ae/∂t (finite diff.)"] - df["Ca"] + df["Ce"] - df["BAe"]
        df["RKe"] = df["∂Ke/∂t (finite diff.)"] - df["Ce"] + df["Ck"] - df["BKe"]

    except Exception as e:
        app_logger.error(f"Error in calc_residuals: {e}")
        raise

    app_logger.debug("Done.")
    return df



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    box_data.py                                        :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/01/31 20:15:59 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 00:19:08 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Create an object that will store, whithin a bounding box specified by the user,
all meteorological data required by the Lorenz Energy Cycle computations.

The program uses xarray built-in function for integrations. For using it, it
would be necessary to sort the data from south to north and from west to east,
but for saving up memory, it was using a negative sign in front of the
integrations for the latitude coordinate.

Also, it was created a dimension for the radians of the latitude and longitude,
for the integration

Suffixes used:
    ZA = Zonal average (whithin the dfined box)
    AA = Area average (whithin the dfined box)
    ZE = zonal eddy (departure from zonal average)
    AE = area eddy (zonal departures from area averages)


Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com

@author: daniloceano
"""

import argparse

import numpy as np
import pandas as pd
import xarray as xr
from metpy.constants import g
from metpy.units import units

from ..utils.calc_averages import CalcAreaAverage, CalcZonalAverage
from ..utils.thermodynamics import AdiabaticHEating, StaticStability


class BoxData:
    """
    A class to store and process meteorological data within a specified bounding box
    for Lorenz Energy Cycle computations.

    Attributes:
        data (xr.Dataset): Dataset containing meteorological data.
        variable_list_df (pd.DataFrame): DataFrame mapping variables to dataset fields.
        western_limit (float): Western boundary of the bounding box.
        eastern_limit (float): Eastern boundary of the bounding box.
        southern_limit (float): Southern boundary of the bounding box.
        northern_limit (float): Northern boundary of the bounding box.
        args (argparse.Namespace): Arguments for additional processing controls.
        results_subdirectory (str): Directory path for output data.
        dTdt (xr.DataArray, optional): Dataset containing temperature gradient data.

    Methods:
        __init__: Initializes the BoxData object with the given parameters.
    """

    def __init__(
        self,
        data: xr.Dataset,
        variable_list_df: pd.DataFrame,
        western_limit: float,
        eastern_limit: float,
        southern_limit: float,
        northern_limit: float,
        args: argparse.Namespace,
        results_subdirectory: str,
        results_subdirectory_vertical_levels: str,
        dTdt: xr.DataArray = None,
    ):
        """Initialize the BoxData object with meteorological data and computation parameters."""
        self._initialize_indices(variable_list_df, data)
        self.args = args
        self.results_subdirectory = results_subdirectory
        self.results_subdirectory_vertical_levels = results_subdirectory_vertical_levels

        self.dx = float(data[self.LatIndexer][1] - data[self.LatIndexer][0])

        # Set domain limits and calculate lengths for averaging
        self._set_domain_limits(
            data, western_limit, eastern_limit, southern_limit, northern_limit
        )

        # Extract and process meteorological data within the defined box
        self._process_meteorological_data(data, variable_list_df, dTdt)

    def _initialize_indices(self, variable_list_df, data):
        """Initialize indices from the variable list DataFrame."""
        self.LonIndexer = variable_list_df.loc["Longitude"]["Variable"]
        self.LatIndexer = variable_list_df.loc["Latitude"]["Variable"]
        self.TimeName = variable_list_df.loc["Time"]["Variable"]
        self.VerticalCoordIndexer = variable_list_df.loc["Vertical Level"]["Variable"]
        self.PressureData = data[self.VerticalCoordIndexer] * units("Pa")

    def _set_domain_limits(
        self, data, western_limit, eastern_limit, southern_limit, northern_limit
    ):
        """Set domain limits for the bounding box and calculate lengths for averaging."""
        self.western_limit = self._select_nearest(data, self.LonIndexer, western_limit)
        self.eastern_limit = self._select_nearest(data, self.LonIndexer, eastern_limit)
        self.southern_limit = self._select_nearest(
            data, self.LatIndexer, southern_limit
        )
        self.northern_limit = self._select_nearest(
            data, self.LatIndexer, northern_limit
        )

        self.xlength = self.eastern_limit["rlons"] - self.western_limit["rlons"]
        self.ylength = np.sin(self.northern_limit["rlats"]) - np.sin(
            self.southern_limit["rlats"]
        )

    def _select_nearest(self, data, indexer, value):
        """Select the data point nearest to the specified value."""
        return data[indexer].sel({indexer: value}, method="nearest")

    def _process_meteorological_data(self, data, variable_list_df, dTdt):
        """Extract and process meteorological data within the defined box."""
        # Extract and process temperature data
        self._process_temperature_data(data, variable_list_df)

        # Extract and process wind component data
        self._process_wind_data(data, variable_list_df)

        # Extract and process frictional terms
        self._process_friction_terms(data, variable_list_df, self.args)

        # Extract and process vertical velocity (omega)
        self._process_omega(data, variable_list_df)

        # Extract and process geopotential
        self._process_geopotential_data(data, variable_list_df)

        # Extract and process static stability parameter
        self._process_adiabatic_heating_and_stability(dTdt)

    def _process_temperature_data(self, data, variable_list_df):
        """Extract and process temperature data."""
        self.tair = self._extract_data(data, variable_list_df, "Air Temperature", "K")
        self.tair_ZA = CalcZonalAverage(self.tair, self.xlength)
        self.tair_AA = CalcAreaAverage(self.tair_ZA, self.ylength)
        self.tair_ZE = self.tair - self.tair_ZA
        self.tair_AE = self.tair_ZA - self.tair_AA

    def _process_wind_data(self, data, variable_list_df):
        """Extract and process wind component data."""
        self.u = self._extract_data(
            data, variable_list_df, "Eastward Wind Component", "m/s"
        )
        self.u_ZA = CalcZonalAverage(self.u, self.xlength)
        self.u_AA = CalcAreaAverage(self.u_ZA, self.ylength)
        self.u_ZE = self.u - self.u_ZA
        self.u_AE = self.u_ZA - self.u_AA

        self.v = self._extract_data(
            data, variable_list_df, "Northward Wind Component", "m/s"
        )
        self.v_ZA = CalcZonalAverage(self.v, self.xlength)
        self.v_AA = CalcAreaAverage(self.v_ZA, self.ylength)
        self.v_ZE = self.v - self.v_ZA
        self.v_AE = self.v_ZA - self.v_AA

    def _process_friction_terms(self, data, variable_list_df, args):
        """Extract and process friction terms."""
        if args.residuals:
            self.ust = self.tair * np.nan
            self.vst = self.tair * np.nan

        else:
            self.ust = self._extract_data(
                data, variable_list_df, "Friction Velocity", "m/s"
            )
            self.vst = self._extract_data(
                data, variable_list_df, "Friction Velocity", "m/s"
            )

        self.ust_ZA = CalcZonalAverage(self.ust, self.xlength)
        self.ust_AA = CalcAreaAverage(self.ust_ZA, self.ylength)
        self.ust_ZE = self.ust - self.ust_ZA
        self.ust_AE = self.ust_ZA - self.ust_AA

        self.vst_ZA = CalcZonalAverage(self.vst, self.xlength)
        self.vst_AA = CalcAreaAverage(self.vst_ZA, self.ylength)
        self.vst_ZE = self.vst - self.vst_ZA
        self.vst_AE = self.vst_ZA - self.vst_AA

    def _process_omega(self, data, variable_list_df):
        """Extract and process vertical velocity."""
        self.omega = self._extract_data(
            data, variable_list_df, "Omega Velocity", "Pa/s"
        )
        self.omega_ZA = CalcZonalAverage(self.omega, self.xlength)
        self.omega_AA = CalcAreaAverage(self.omega_ZA, self.ylength)
        self.omega_ZE = self.omega - self.omega_ZA
        self.omega_AE = self.omega_ZA - self.omega_AA

    def _process_geopotential_data(self, data, variable_list_df):
        """Extract and process geopotential or geopotential height data."""
        if "Geopotential" in variable_list_df.index:
            self.geopt = self._extract_data(
                data, variable_list_df, "Geopotential", "m**2/s**2"
            )
        else:
            self.geopt = self._extract_and_convert_geopotential_height(
                data, variable_list_df
            )

        self.geopt_ZA = CalcZonalAverage(self.geopt, self.xlength)
        self.geopt_AA = CalcAreaAverage(self.geopt_ZA, self.ylength)
        self.geopt_ZE = self.geopt - self.geopt_ZA
        self.geopt_AE = self.geopt_ZA - self.geopt_AA

    def _extract_and_convert_geopotential_height(self, data, variable_list_df):
        """Extract geopotential height data and convert to geopotential."""
        geopt_height = self._extract_data(
            data,
            variable_list_df,
            "Geopotential Height",
            variable_list_df.loc["Geopotential Height"]["Units"],
        )
        return (geopt_height * g).metpy.convert_units("m**2/s**2")

    def _process_adiabatic_heating_and_stability(self, dTdt):
        """Process adiabatic heating and static stability parameter."""
        if self.args.fixed:
            self.Q = AdiabaticHEating(
                self.tair,
                self.PressureData,
                self.omega,
                self.u,
                self.v,
                self.VerticalCoordIndexer,
                self.LatIndexer,
                self.LonIndexer,
                self.TimeName,
            ).sel(
                **{
                    self.LatIndexer: slice(self.southern_limit, self.northern_limit),
                    self.LonIndexer: slice(self.western_limit, self.eastern_limit),
                }
            )
        elif self.args.track or self.args.choose:
            self.dTdt = dTdt
            self.Q = AdiabaticHEating(
                self.tair,
                self.PressureData,
                self.omega,
                self.u,
                self.v,
                self.VerticalCoordIndexer,
                self.LatIndexer,
                self.LonIndexer,
                self.TimeName,
                self.dTdt,
            ).sel(
                **{
                    self.LatIndexer: slice(self.southern_limit, self.northern_limit),
                    self.LonIndexer: slice(self.western_limit, self.eastern_limit),
                }
            )
        else:
            print("Adiabatic heating computation skipped. Check flags!")

        self.Q_ZA = CalcZonalAverage(self.Q, self.xlength)
        self.Q_AA = CalcAreaAverage(self.Q_ZA, self.ylength)
        self.Q_ZE = self.Q - self.Q_ZA
        self.Q_AE = self.Q_ZA - self.Q_AA

        self.sigma_AA = StaticStability(
            self.tair,
            self.PressureData,
            self.VerticalCoordIndexer,
            self.xlength,
            self.ylength,
        )

    def _extract_data(self, data, variable_list_df, variable_name, unit):
        """Extract data for a specific variable and convert to the specified unit."""
        var_key = variable_list_df.loc[variable_name]["Variable"]
        unit_to_convert = units(variable_list_df.loc[variable_name]["Units"])
        return (
            (data[var_key] * unit_to_convert)
            .metpy.convert_units(unit)
            .sel(
                **{
                    self.LatIndexer: slice(self.southern_limit, self.northern_limit),
                    self.LonIndexer: slice(self.western_limit, self.eastern_limit),
                }
            )
        )



import os

from setuptools import find_packages, setup

# Utility function to read the requirements.txt file


def read_requirements():
    requirements_path = os.path.join(os.path.dirname(__file__), 'requirements.txt')
    with open(requirements_path, 'r') as f:
        return f.read().splitlines()


setup(
    name='LorenzCycleToolkit',
    version='1.0.9',
    packages=find_packages(),
    include_package_data=True,
    install_requires=read_requirements(),
    entry_points={
        'console_scripts': [
            'lorenzcycletoolkit = lorenzcycletoolkit:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
)



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    plot_LPS.py                                        :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/06/14 16:32:27 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/14 19:36:12 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import logging
import os
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
from lorenz_phase_space.phase_diagrams import Visualizer


def create_and_save_plot(
    dataframe: pd.DataFrame,
    title: str,
    zoom: bool,
    plot_type: str,
    figures_dir: Path,
    app_logger: logging.Logger,
):

    x_min, x_max = dataframe["Ck"].min(), dataframe["Ck"].max()
    y_min, y_max = dataframe["Ca"].min(), dataframe["Ca"].max()
    color_min, color_max = dataframe["Ge"].min(), dataframe["Ge"].max()
    size_min, size_max = dataframe["Ke"].min(), dataframe["Ke"].max()

    # Initialize the Visualizer with or without zoom
    lps = Visualizer(
        LPS_type="mixed",
        zoom=zoom,
        x_limits=[x_min, x_max],
        y_limits=[y_min, y_max],
        color_limits=[color_min, color_max],
        marker_limits=[size_min, size_max],
    )

    # Plot the data
    lps.plot_data(
        x_axis=dataframe["Ck"].values,
        y_axis=dataframe["Ca"].values,
        marker_color=dataframe["Ge"].values,
        marker_size=dataframe["Ke"].values,
    )

    plt.text(
        0,
        1,
        title,
        va="bottom",
        ha="left",
        fontsize=14,
        fontweight="bold",
        color="gray",
        transform=plt.gca().transAxes,
    )

    # Construct filename and save the plot
    suffix = f"_{plot_type}_zoom" if zoom else f"_{plot_type}"
    file_path = os.path.join(figures_dir, f"LPS{suffix}.png")
    plt.savefig(file_path, dpi=300)
    plt.close()  # Close the plot to free memory
    app_logger.info(f"LPS plot saved to {file_path}")


def plot_LPS(dataframe, infile, results_subdirectory, figures_directory, app_logger):
    app_logger.info("Plotting Lorenz Phase Space...")
    infile_name = Path(infile).stem

    figures_LPS_directory = f"{figures_directory}/LPS"
    os.makedirs(figures_LPS_directory, exist_ok=True)

    # Calculate the time difference between two consecutive timestamps
    time_difference = dataframe.index[1] - dataframe.index[0]

    # Convert the time difference to total seconds and then to hours
    dt_hours = int(time_difference.total_seconds() / 3600)

    periods_file = f"{results_subdirectory}/periods.csv"
    try:
        periods_df = pd.read_csv(
            periods_file, parse_dates=["start", "end"], index_col=0
        )
    except FileNotFoundError:
        app_logger.error("Periods file not found.")
        return
    except Exception as e:
        app_logger.error(f"Error while reading periods file: {e}")
        return

    # Initialize an empty DataFrame to store period means
    period_means_df = pd.DataFrame()

    # Iterate through each period and calculate means
    for period_name, row in periods_df.iterrows():
        start, end = row["start"], row["end"]
        df_period = dataframe.loc[start:end]

        # Check if the period DataFrame is not empty
        if not df_period.empty:
            # Calculate mean for the period
            period_mean = df_period.mean()
            # Add the mean to the period_means_df DataFrame
            period_means_df = pd.concat(
                [period_means_df, pd.DataFrame(period_mean).transpose()],
                ignore_index=True,
            )
            period_means_df.index = period_means_df.index.map(
                lambda x: period_name if x == period_means_df.index[-1] else x
            )
        else:
            app_logger.warning(f"No data available for the period: {period_name}")

    try:
        title, datasource = infile_name.split("_")
    except ValueError:
        title, datasource = infile_name, "unknown"

    figure_title = (
        f"System: {title}\n"
        f"Datasource: {datasource}\n"
        f"Start: {dataframe.index[0].strftime('%Y-%m-%d %HZ')}\n"
        f"End: {dataframe.index[-1].strftime('%Y-%m-%d %HZ')}"
    )

    for zoom in [False, True]:
        create_and_save_plot(
            dataframe,
            figure_title,
            zoom,
            f"{dt_hours}h",
            figures_LPS_directory,
            app_logger,
        )
        create_and_save_plot(
            period_means_df,
            figure_title,
            zoom,
            "periods",
            figures_LPS_directory,
            app_logger,
        )
        df_daily = dataframe.resample("1D").mean()
        create_and_save_plot(
            df_daily, figure_title, zoom, "1d", figures_LPS_directory, app_logger
        )



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    map_track.py                                       :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/07/05 19:48:17 by daniloceano       #+#    #+#              #
#    Updated: 2024/01/04 10:40:24 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com

"""

import os
from bisect import bisect_left

import cartopy.crs as ccrs
import cmocean
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np

import src.plots.utils as utils
from src.plots.utils import (map_borders, read_results, read_track,
                             setup_gridlines, setup_map)

COLORS = ["#BF3D3B", "#3B95BF"]


def set_plot_legend(fig, labels, title, msizes):
    """
    Set the plot legend with custom scatter plot elements.

    Parameters:
    fig (matplotlib.figure.Figure): The figure object to which the legend is added.
    labelx (float): The x-coordinate for the legend placement.
    labels (list of str): Labels for each category in the legend.
    title (str): Title of the legend.
    msizes (list of int): Marker sizes corresponding to each label.
    """
    legend_elements = [
        plt.scatter([], [], c=utils.TEXT_COLOR, s=size, label=label)
        for label, size in zip(labels, msizes)
    ]
    leg = fig.legend(
        handles=legend_elements,
        ncol=1,
        frameon=False,
        fontsize=utils.LEGEND_FONT_SIZE,
        handlelength=0.3,
        handleheight=4,
        borderpad=1.5,
        scatteryoffsets=[0.1],
        framealpha=1,
        handletextpad=1.5,
        title=title,
        scatterpoints=1,
        loc="center left",
        bbox_to_anchor=(0.8, 0.2),
        labelcolor=utils.TEXT_COLOR,
    )
    leg._legend_box.align = "center"
    plt.setp(
        leg.get_title(),
        color=utils.TEXT_COLOR,
        ha="center",
        fontsize=utils.AXIS_LABEL_FONT_SIZE,
    )


def map_track(results_file_path, trackfile, figures_subdirectory, app_logger=False):
    """
    Generate a map plot of track data with scatter plot elements.

    Parameters:
    results_file_path (str): Path to the results CSV file.
    trackfile (str): Path to the track CSV file.
    figures_subdirectory (str): Directory path to save the generated plot.
    """
    app_logger.info("Plotting track...") if app_logger else print("Plotting track...")

    track = read_track(trackfile, app_logger)
    df = read_results(results_file_path, app_logger)

    # Extract track latitudes and longitudes
    try:
        track_lon = track["Lon"]
        track_lat = track["Lat"]
    except KeyError:
        (
            app_logger.error("Error: Lon and Lat columns not found in track CSV file.")
            if app_logger
            else print("Error: Lon and Lat columns not found in track CSV file.")
        )
        return

    plt.close("all")
    fig, ax = setup_figure(track_lon, track_lat)
    plot_track_data(fig, ax, df, track)

    figure_file = os.path.join(figures_subdirectory, "track.png")
    plt.savefig(figure_file, bbox_inches="tight")
    (
        app_logger.info(f"Track plot saved to {figure_file}")
        if app_logger
        else print(f"Track plot saved to {figure_file}")
    )


def setup_figure(track_lon, track_lat):
    min_lon, max_lon, min_lat, max_lat = (
        track_lon.min(),
        track_lon.max(),
        track_lat.min(),
        track_lat.max(),
    )
    lenx, leny = max_lon - min_lon, max_lat - min_lat
    fig_size = (12, 9) if lenx <= leny else (14, 9)
    fig = plt.figure(figsize=fig_size)
    ax = fig.add_axes(
        [0.05, 0.05, 0.9, 0.9], projection=ccrs.PlateCarree(), frameon=True
    )
    ax.set_extent(
        [min_lon - 10, max_lon + 10, max_lat + 10, min_lat - 10], crs=ccrs.PlateCarree()
    )
    setup_map(ax)
    setup_gridlines(ax)
    map_borders(ax)
    return fig, ax


def plot_track_data(fig, ax, df, track, app_logger=False):
    # Define sizes and intervals for scatter plot markers
    num_quantiles = 5  # Number of quantiles (or size categories)
    msizes = np.linspace(100, 1000, num_quantiles)

    # Extract Ke and Ae values
    try:
        Ke = df["Ke"]
        Ae = df["Ae"]
    except KeyError:
        (
            app_logger.error("Error: Ke and Ae columns not found in results CSV file.")
            if app_logger
            else print("Error: Ke and Ae columns not found in results CSV file.")
        )
        return

    # Extract track latitudes and longitudes
    track_lon, track_lat = track["Lon"], track["Lat"]

    # Scale down Ke and Ae by 10^5
    Ke_scaled = Ke / 1e5
    Ae_scaled = Ae / 1e5

    # Calculate quantile values for scaled Ke
    quantile_values = Ke_scaled.quantile(
        [i / (num_quantiles - 1) for i in np.arange(1, num_quantiles)]
    ).tolist()

    # Assign sizes based on quantiles
    def assign_Ke_size(value):
        index = bisect_left(quantile_values, value)
        # Ensure index is within the range of msizes
        return msizes[min(index, len(msizes) - 1)]

    df["sizes"] = Ke_scaled.apply(assign_Ke_size)

    # Create labels for each quantile interval with rounded values
    # Label for values below the first quantile
    labels = [f"< {quantile_values[0]:.2f}"]
    for i in range(len(quantile_values) - 1):
        labels.append(f"{quantile_values[i]:.2f} - {quantile_values[i + 1]:.2f}")
    # Label for values above the last quantile
    labels.append(f"> {quantile_values[-1]:.2f}")

    # Add label for the max range
    labels[-1] = f"> {quantile_values[-1]:.2f}"
    title = r"Eddy Kinect Energy" + "\n" r"(Ke) [x $10^{5}$ J m$^{-2}$]"
    set_plot_legend(fig, labels, title, msizes)

    # Plotting the track and scatter plot
    ax.plot(track_lon, track_lat, c=utils.TEXT_COLOR)
    norm = colors.TwoSlopeNorm(
        vmin=min(Ae_scaled), vcenter=np.mean(Ae_scaled), vmax=max(Ae_scaled)
    )

    # Adjust DataFrame if necessary
    if len(track) < len(df):
        df = df[df["Datetime"].isin(track.index)]
    elif len(df) < len(track):
        track = track[track.index.isin(df["Datetime"])]

    # Scatter plot for tracking data
    dots = ax.scatter(
        track_lon,
        track_lat,
        c=Ae_scaled,
        cmap=cmocean.cm.matter,
        s=df["sizes"],
        zorder=100,
        edgecolors=utils.MARKER_EDGE_COLOR,
        norm=norm,
    )

    # Colorbar setup
    cax = fig.add_axes(
        [
            ax.get_position().x1 + 0.02,
            ax.get_position().y0 + 0.38,
            0.02,
            ax.get_position().height / 1.74,
        ]
    )
    cbar = plt.colorbar(dots, extend="both", cax=cax)
    cbar.ax.set_ylabel(
        "Eddy Potential Energy (Ae) [x $10^{5}$ " + r" $J\,m^{-2}$" + "]",
        rotation=270,
        fontsize=utils.AXIS_LABEL_FONT_SIZE,
        verticalalignment="bottom",
        color=utils.TEXT_COLOR,
        labelpad=15,
    )

    # Mark the beginning and end of the system
    start = [track_lon.iloc[0], track_lat.iloc[0]]
    end = [track_lon.iloc[-1], track_lat.iloc[-1]]
    ax.text(
        *start,
        "A",
        zorder=101,
        fontsize=24,
        horizontalalignment="center",
        verticalalignment="center",
    )
    ax.text(
        *end,
        "Z",
        zorder=101,
        fontsize=24,
        horizontalalignment="center",
        verticalalignment="center",
    )


if __name__ == "__main__":
    results_file_path = "samples/sample_results.csv"
    track_file = "samples/sample_track.csv"
    figures_subdirectory = "samples/Figures/"
    os.makedirs(figures_subdirectory, exist_ok=True)
    map_track(results_file_path, track_file, figures_subdirectory)
